"""
Functions for calculating brain connectivity metrics.

Uses naming conventions adopted from the Brain Connectivity
Toolbox (https://sites.google.com/site/bctnet/).
"""

import itertools
import numpy as np
import scipy
from scipy.stats import ttest_ind
from scipy.sparse.csgraph import shortest_path

from .. import has_numba

if has_numba:
    from numba import njit
from .metrics_utils import _fast_binarize


def degrees_und(W):
    """
    Compute the degree of each node in `W`.

    Parameters
    ----------
    W : (N, N) array_like
        Unweighted, undirected connection weight array.
        Weighted array will be binarized prior to calculation.
        Directedness will be ignored (out degree / row sum taken).

    Returns
    -------
    deg : (N,) numpy.ndarray
        Degree of each node in `W`
    """
    return np.sum(_fast_binarize(W), axis=0)


def degrees_dir(W):
    """
    Compute the in degree and out degree of each node in `W`.

    Parameters
    ----------
    W : (N, N) array_like
        Unweighted, directed connection weight array.
        Weighted array will be binarized prior to calculation.

    Returns
    -------
    deg_in : (N,) numpy.ndarray
        In-degree (column sum) of each node in `W`
    deg_out : (N,) numpy.ndarray
        Out-degree (row sum) of each node in `W`
    deg : (N,) numpy.ndarray
        Degree (in-degree + out-degree) of each node in `W`
    """
    W_bin = _fast_binarize(W)
    deg_in = np.sum(W_bin, axis=0)
    deg_out = np.sum(W_bin, axis=1)
    deg = deg_in + deg_out
    return deg_in, deg_out, deg


def distance_wei_floyd(D):
    """
    Compute the all-pairs shortest path length using Floyd-Warshall algorithm.

    Parameters
    ----------
    D : (N, N) array_like
        Connection length or distance matrix.
        Please do the weight-to-distance beforehand.

    Returns
    -------
    spl_mat : (N, N) array_like
        Shortest path length matrix
    p_mat : (N, N) array_like
        Predecessor matrix returned from `scipy.sparse.csgraph.shortest_path`

    Notes
    -----
    This function is a wrapper for `scipy.sparse.csgraph.shortest_path`.
    There may be more than one shortest path between two nodes, and we
    only return the first one found by the algorithm.

    References
    ----------
    .. [1] Floyd, R. W. (1962). Algorithm 97: shortest path. Communications of
       the ACM, 5(6), 345.
    .. [2] Roy, B. (1959). Transitivite et connexite. Comptes Rendus
       Hebdomadaires Des Seances De L Academie Des Sciences, 249(2), 216-218.
    .. [3] Warshall, S. (1962). A theorem on boolean matrices. Journal of the
       ACM (JACM), 9(1), 11-12.
    .. [4] https://en.wikipedia.org/wiki/Floyd%E2%80%93Warshall_algorithm

    See Also
    --------
    netneurotools.metrics.retrieve_shortest_path
    """
    spl_mat, p_mat = shortest_path(
        D,
        method="FW",
        directed=True,
        return_predecessors=True,
        unweighted=False,
        overwrite=False,
    )
    return spl_mat, p_mat


def retrieve_shortest_path(s, t, p_mat):
    """
    Return the shortest paths between two nodes.

    Parameters
    ----------
    s : int
        Source node
    t : int
        Target node
    p_mat : (N, N) array_like
        Predecessor matrix returned from `distance_wei_floyd`

    Returns
    -------
    path : list of int
        List of nodes in the shortest path from `s` to `t`. If no path
        exists, returns `[-1]`.

    See Also
    --------
    netneurotools.metrics.distance_wei_floyd
    """
    if p_mat[s, t] == -9999:
        return [-1]
    path = [t]
    while path[-1] != s:
        t = p_mat[s, t]
        path.append(t)
    return path[::-1]


if has_numba:
    retrieve_shortest_path = njit(retrieve_shortest_path)


def navigation_wu(nav_dist_mat, sc_mat):
    """
    Compute network navigation.

    Parameters
    ----------
    nav_dist_mat : (N, N) array_like
        Connection length/distance matrix.
    sc_mat : (N, N) array_like
        Structural connectivity matrix, only used to get connectedness.

    Returns
    -------
    nav_sr : float
        Overall navigation success rate
    nav_sr_node : list of float
        Nodal navigation success rate
    nav_path_len : (N, N) array_like
        Navigation path length matrix, infinite if no path exists.
    nav_path_hop : (N, N) array_like
        Navigation path hop matrix, infinite if no path exists.
    nav_paths : list
        List of tuples containing source node, target node, path length,
        path hops, and the full path.

    References
    ----------
    Seguin, C., Van Den Heuvel, M. P., & Zalesky, A. (2018). Navigation
    of brain networks. Proceedings of the National Academy of Sciences,
    115(24), 6297-6302.

    Notes
    -----
    Euclidean distance between nodes are usually used for `nav_dist_mat`.
    Distances returned from this function are also calculated from
    `nav_dist_mat`.
    Use :meth:`netneurotools.metrics.get_navigation_path_length`
    to get path length in other metrics.

    See Also
    --------
    netneurotools.metrics.get_navigation_path_length
    """
    nav_paths = []  # (source, target, distance, hops, path)
    # navigate to the node that is closest to target
    for src in range(len(nav_dist_mat)):
        for tar in range(len(nav_dist_mat)):
            curr_pos = src
            curr_path = [src]
            curr_dist = 0
            while curr_pos != tar:
                neig = np.where(sc_mat[curr_pos, :] != 0)[0]
                if len(neig) == 0:  # not connected
                    curr_path = []
                    curr_dist = np.inf
                    break
                neig_dist_to_tar = nav_dist_mat[neig, tar]
                min_dist_idx = np.argmin(neig_dist_to_tar)

                new_pos = neig[min_dist_idx]
                # Assume it is connected, and only testing for loops.
                # if isempty(next_node)
                # || next_node == last_node
                # || pl_bin > max_hops
                if new_pos in curr_path:
                    curr_path = []
                    curr_dist = np.inf
                    break
                else:
                    curr_path.append(new_pos)
                    curr_dist += nav_dist_mat[curr_pos, new_pos]
                    curr_pos = new_pos
            nav_paths.append((src, tar, curr_dist, len(curr_path) - 1, curr_path))

    nav_sr = len([_ for _ in nav_paths if _[3] != -1]) / len(nav_paths)

    nav_sr_node = []
    for _, g in itertools.groupby(
        sorted(nav_paths, key=lambda x: x[0]), key=lambda x: x[0]
    ):
        curr_path = list(g)
        nav_sr_node.append(len([_ for _ in curr_path if _[3] != -1]) / len(curr_path))

    nav_path_len = np.zeros_like(nav_dist_mat)
    nav_path_hop = np.zeros_like(nav_dist_mat)
    for nav_item in nav_paths:
        i, j, length, hop, _ = nav_item
        if hop != -1:
            nav_path_len[i, j] = length
            nav_path_hop[i, j] = hop
        else:
            nav_path_len[i, j] = np.inf
            nav_path_hop[i, j] = np.inf

    return nav_sr, nav_sr_node, nav_path_len, nav_path_hop, nav_paths


def get_navigation_path_length(nav_paths, alt_dist_mat):
    """
    Get navigation path length from navigation results.

    Parameters
    ----------
    nav_paths : list
        Return from netneurotools.metrics.navigation_wu
    alt_dist_mat : (N, N) array_like
        Alternative distance matrix, e.g. geodesic distance.

    Returns
    -------
    nav_path_len : (N, N) array_like
        Navigation path length matrix, in the alternative distance metric.

    Notes
    -----
    Following the original BCT function.
    `pl_wei = get_navigation_path_length(nav_paths, L)`
    L is strength-to-length remapping of the connection weight matrix.
    `pl_dis = get_navigation_path_length(nav_paths, D)`
    D is Euclidean distance between node centroids.

    See Also
    --------
    netneurotools.metrics.navigation_wu
    """
    nav_path_len = np.zeros_like(alt_dist_mat)
    for nav_item in nav_paths:
        i, j, _, hop, path = nav_item
        if hop != -1:
            nav_path_len[i, j] = np.sum(
                [alt_dist_mat[path[_], path[_ + 1]] for _ in range(hop)]
            )
        else:
            nav_path_len[i, j] = np.inf
    return nav_path_len


def communicability_bin(adjacency, normalize=False):
    """
    Compute the communicability of pairs of nodes in `adjacency`.

    Parameters
    ----------
    adjacency : (N, N) array_like
        Unweighted, direct/undirected connection weight/length array
    normalize : bool, optional
        Whether to normalize `adjacency` by largest eigenvalue prior to
        calculation of communicability metric. Default: False

    Returns
    -------
    comm : (N, N) numpy.ndarray
        Symmetric array representing communicability of nodes {i, j}

    References
    ----------
    Estrada, E., & Hatano, N. (2008). Communicability in complex networks.
    Physical Review E, 77(3), 036111.

    Examples
    --------
    >>> from netneurotools import metrics

    >>> A = np.array([[1, 0, 1], [0, 1, 1], [1, 0, 1]])
    >>> Q = metrics.communicability_bin(A)
    >>> Q
    array([[4.19452805, 0.        , 3.19452805],
           [1.47624622, 2.71828183, 3.19452805],
           [3.19452805, 0.        , 4.19452805]])
    """
    if not np.any(np.logical_or(adjacency == 0, adjacency == 1)):
        raise ValueError("Provided adjancecy matrix must be unweighted.")

    # normalize by largest eigenvalue to prevent communicability metric from
    # "blowing up"
    if normalize:
        norm = np.linalg.eigvals(adjacency).max()
        adjacency = adjacency / norm

    return scipy.sparse.linalg.expm(adjacency)


def communicability_wei(adjacency):
    """
    Compute the communicability of pairs of nodes in `adjacency`.

    Parameters
    ----------
    adjacency : (N, N) array_like
        Weighted, direct/undirected connection weight/length array

    Returns
    -------
    cmc : (N, N) numpy.ndarray
        Symmetric array representing communicability of nodes {i, j}

    References
    ----------
    Crofts, J. J., & Higham, D. J. (2009). A weighted communicability measure
    applied to complex brain networks. Journal of the Royal Society Interface,
    6(33), 411-414.

    Examples
    --------
    >>> from netneurotools import metrics

    >>> A = np.array([[2, 0, 3], [0, 2, 1], [0.5, 0, 1]])
    >>> Q = metrics.communicability_wei(A)
    >>> Q
    array([[0.        , 0.        , 1.93581903],
           [0.07810379, 0.        , 0.94712177],
           [0.32263651, 0.        , 0.        ]])
    """
    # negative square root of nodal degrees
    row_sum = adjacency.sum(1)
    neg_sqrt = np.power(row_sum, -0.5)
    square_sqrt = np.diag(neg_sqrt)

    # normalize input matrix
    for_expm = square_sqrt @ adjacency @ square_sqrt

    # calculate matrix exponential of normalized matrix
    cmc = scipy.sparse.linalg.expm(for_expm)
    cmc[np.diag_indices_from(cmc)] = 0

    return cmc


def path_transitivity(D):
    """
    Calculate path transitivity.

    This function implements path transitivity, calculating the density of
    local detours (triangles) that are available along the shortest paths
    between all pairs of nodes.

    This function is adapted and optimized from the Brain Connectivity Toolbox.

    .. warning::
       Test before use.

    Parameters
    ----------
    D : (N, N) ndarray
        Weight or connection length matrix. Length matrix is recommended and
        transform should have been applied.

    Returns
    -------
    T_mat : (N, N) ndarray
        Path transitivity matrix

    References
    ----------
    .. [1] Goñi, J., Van Den Heuvel, M. P., Avena-Koenigsberger,
       A., Velez de Mendizabal, N., Betzel, R. F., Griffa, A., ... &
       Sporns, O. (2014). Resting-brain functional connectivity predicted
       by analytic measures of network communication. Proceedings of the
       National Academy of Sciences, 111(2), 833-838.
    """
    n = len(D)
    m = np.zeros((n, n))
    T_mat = np.zeros((n, n))

    deg_wu = np.sum(D, axis=0)

    for i in range(n - 1):
        for j in range(i + 1, n):
            sig_and = np.logical_and(D[i, :], D[j, :])
            m[i, j] = np.dot(D[i, :] + D[j, :], sig_and) / (
                deg_wu[i] + deg_wu[j] - 2 * D[i, j]
            )
    m += m.transpose()

    _, p_mat = distance_wei_floyd(D)

    for i in range(n - 1):
        for j in range(i + 1, n):
            path = retrieve_shortest_path(i, j, p_mat)
            K = len(path)
            T_mat[i, j] = (
                2
                * sum([m[i, j] for i, j in itertools.combinations(path, 2)])
                / (K * (K - 1))
            )
    T_mat += T_mat.transpose()

    return T_mat


def search_information(W, D, has_memory=False):
    """
    Calculate search information.

    This function implements search information, computes the amount
    of information (measured in bits) that a random walker needs to
    follow the shortest path between a given pair of nodes.

    This function is adapted and optimized from the Brain Connectivity Toolbox.

    .. warning::
       Test before use.

    Parameters
    ----------
    W : (N, N) ndarray
        Weighted/unweighted, directed/undirected connection weight matrix.
    D : (N, N) ndarray
        Weighted/unweighted, directed/undirected connection length or
        distance matrix. Please do the weight-to-distance beforehand.
    has_memory : bool, optional
        Memory for random walker, Default: False

    Returns
    -------
    SI : (N, N) ndarray
        Pairwise search information matrix. The diagonal is set to NaN.
        Edges without a valid shortest path are set to np.inf.
        It is not guaranteed to be symmetric even if input is symmetric.

    References
    ----------
    .. [1] Rosvall, M., Trusina, A., Minnhagen, P., & Sneppen, K. (2005).
       Networks and cities: An information perspective. Physical Review
       Letters, 94(2), 028701.
    .. [2] Goñi, J., Van Den Heuvel, M. P., Avena-Koenigsberger,
       A., Velez de Mendizabal, N., Betzel, R. F., Griffa, A., ... &
       Sporns, O. (2014). Resting-brain functional connectivity predicted
       by analytic measures of network communication. Proceedings of the
       National Academy of Sciences, 111(2), 833-838.
    """
    N = len(W)

    is_sym = True if np.allclose(W, W.T) else False

    T = W / np.sum(W, axis=1)[:, None]
    _, p_mat = distance_wei_floyd(D)

    SI = np.zeros((N, N))

    if is_sym:  # symmetric case
        for i in range(N):
            for j in range(i + 1, N):
                path = retrieve_shortest_path(i, j, p_mat)
                if path[0] != -1:  # no path, depends on retrieve_shortest_path
                    if has_memory:
                        pr_step_ff = [0] + [
                            T[i, j] for i, j in zip(path[:-1], path[1:])
                        ]
                        pr_step_bk = [T[i, j] for i, j in zip(path[1:], path[:-1])] + [
                            0
                        ]
                        pr_step_ff = [
                            i / (1 - j) for i, j in zip(pr_step_ff[1:], pr_step_ff[:-1])
                        ]
                        pr_step_bk = [
                            i / (1 - j) for i, j in zip(pr_step_bk[:-1], pr_step_bk[1:])
                        ]
                    else:
                        pr_step_ff = [T[i, j] for i, j in zip(path[:-1], path[1:])]
                        pr_step_bk = [T[i, j] for i, j in zip(path[1:], path[:-1])]
                    SI[i, j] = -np.log2(np.prod(pr_step_ff))
                    SI[j, i] = -np.log2(np.prod(pr_step_bk))
                else:
                    SI[i, j] = np.inf
                    SI[j, i] = np.inf
    else:  # asymmetric case
        for i in range(N):
            for j in range(N):
                if i == j:  # skip self connection
                    continue
                path = retrieve_shortest_path(i, j, p_mat)
                if path[0] != -1:  # no path, depends on retrieve_shortest_path
                    if has_memory:
                        pr_step_ff = [0] + [
                            T[i, j] for i, j in zip(path[:-1], path[1:])
                        ]
                        pr_step_ff = [
                            i / (1 - j) for i, j in zip(pr_step_ff[1:], pr_step_ff[:-1])
                        ]
                    else:
                        pr_step_ff = [T[i, j] for i, j in zip(path[:-1], path[1:])]
                    SI[i, j] = -np.log2(np.prod(pr_step_ff))
                else:
                    SI[i, j] = np.inf

    np.fill_diagonal(SI, np.nan)

    return SI


def mean_first_passage_time(W, tol=1e-3):
    """
    Calculate mean first passage time.

    The first passage time from i to j is the expected number of steps it takes
    a random walker starting at node i to arrive for the first time at node j.
    The mean first passage time is not a symmetric measure: `mfpt(i,j)` may be
    different from `mfpt(j,i)`.

    This function is adapted and optimized from the Brain Connectivity Toolbox.

    .. warning::
       Test before use.

    Parameters
    ----------
    W : (N x N) ndarray
        Weighted/unweighted, direct/undirected connection weight/length array
    tol : float, optional
        Tolerance for eigenvalue of 1. Default: 1e-3

    Returns
    -------
    mfpt : (N x N) ndarray
        Pairwise mean first passage time array

    References
    ----------
    .. [1] Goñi, J., Avena-Koenigsberger, A., Velez de Mendizabal, N.,
       van den Heuvel, M. P., Betzel, R. F., & Sporns, O. (2013). Exploring the
       morphospace of communication efficiency in complex networks. PLoS One,
       8(3), e58070.
    """
    P = W / np.sum(W, axis=1)[:, None]  # transition matrix
    n = len(P)
    D, V = np.linalg.eig(P.T)
    D_minidx = np.argmin(np.abs(D - 1))

    if D[D_minidx] > 1 + tol:
        raise ValueError(
            f"Cannot find eigenvalue of 1. Minimum eigenvalue is larger than {tol}."
        )

    w = V[:, D_minidx][None, :]
    w /= np.sum(w)
    W_prob = np.real(np.repeat(w, n, 0))
    Z = np.linalg.inv(np.eye(n) - P + W_prob)  # fundamental matrix
    mfpt = (np.repeat(np.diag(Z)[None, :], n, 0) - Z) / W_prob
    return mfpt


def diffusion_efficiency(W):
    """
    Calculate diffusion efficiency.

    The diffusion efficiency between nodes i and j is the inverse of the
    mean first passage time from i to j, that is the expected number of
    steps it takes a random walker starting at node i to arrive for the
    first time at node j. Note that the mean first passage time is not a
    symmetric measure -- mfpt(i,j) may be different from mfpt(j,i) -- and
    the pair-wise diffusion efficiency matrix is hence also not symmetric.

    This function is adapted and optimized from the Brain Connectivity Toolbox.

    .. warning::
       Test before use.

    Parameters
    ----------
    W : (N x N) ndarray
        Weighted/unweighted, direct/undirected connection weight/length array

    Returns
    -------
    GE_diff : float
        Global diffusion efficiency
    E_diff : (N x N) ndarray
        Pair-wise diffusion efficiency array

    References
    ----------
    .. [1] Goñi, J., Avena-Koenigsberger, A., Velez de Mendizabal, N.,
       van den Heuvel, M. P., Betzel, R. F., & Sporns, O. (2013). Exploring the
       morphospace of communication efficiency in complex networks. PLoS One,
       8(3), e58070.
    """
    n = W.shape[0]
    mfpt = mean_first_passage_time(W)
    E_diff = np.divide(1, mfpt)
    np.fill_diagonal(E_diff, 0.0)
    GE_diff = np.sum(E_diff) / (n * (n - 1))
    return GE_diff, E_diff


def resource_efficiency_bin(W_bin, lambda_prob=0.5):
    """
    Calculate resource efficiency and shortest-path probability.

    The resource efficiency between nodes i and j is inversly proportional
    to the amount of resources (i.e. number of particles or messages)
    required to ensure with probability 0 < lambda < 1 that at least one of
    them will arrive at node j in exactly SPL steps, where SPL is the
    length of the shortest-path between i and j.

    The shortest-path probability between nodes i and j is the probability
    that a single random walker starting at node i will arrive at node j by
    following (one of) the shortest path(s).

    This function is adapted and optimized from the Brain Connectivity Toolbox.

    .. warning::
       Test before use.

    Parameters
    ----------
    W : (N x N) array_like
        Binary (unweighted) undirected connection matrix.
    lambda_prob : float, optional
        Probability of reaching the target node. Default: 0.5

    Returns
    -------
    E_res : (N x N) ndarray
        Resource efficiency array
    prob_spl : (N x N) ndarray
        Shortest-path probability array

    References
    ----------
    .. [1] Goñi, J., Avena-Koenigsberger, A., Velez de Mendizabal, N.,
       van den Heuvel, M. P., Betzel, R. F., & Sporns, O. (2013). Exploring the
       morphospace of communication efficiency in complex networks. PLoS One,
       8(3), e58070.
    """
    W_bin = _fast_binarize(W_bin)
    if not (0 < lambda_prob < 1):
        raise ValueError("lambda_prob must be between 0 and 1.")

    N = W_bin.shape[0]
    spl_mat, _ = distance_wei_floyd(W_bin)
    spl_mat = spl_mat.astype(int)
    T = W_bin / np.sum(W_bin, axis=1)[:, None]

    L_unique = np.unique(spl_mat)
    prob_spl = np.zeros((N, N))
    z = np.zeros((N, N))

    for L_value in L_unique:
        if L_value == 0:
            continue
        L_locs = spl_mat == L_value
        h_cols = np.where(L_locs)[1]
        h_vec = np.unique(h_cols)

        prob_aux = np.zeros((N, N))
        z_aux = np.zeros((N, N))
        for h_value in h_vec:
            B_h = T.copy()
            B_h[h_value, :] = 0
            B_h[h_value, h_value] = 1
            B_h_L = np.linalg.matrix_power(B_h, L_value)
            prob_aux[:, h_value] = B_h_L[:, h_value]
            z_aux[:, h_value] = np.divide(
                np.ones((N,)) * np.log(1 - lambda_prob), np.log(1 - B_h_L[:, h_value])
            )

        prob_aux[~L_locs] = 0
        prob_spl += prob_aux

        z_aux[~L_locs] = 0
        z += z_aux

    np.fill_diagonal(prob_spl, 0.0)
    z[prob_spl == 1] = 1
    E_res = 1 / z
    np.fill_diagonal(E_res, 0.0)

    return E_res, prob_spl


def flow_graph(W, r=None, t=1):
    """
    Calculate flow graph.

    This function implements flow graph, instantiates a continuous
    time random walk on network. Waiting time for walkers at each
    node are distributed as Poisson with rate parameter r.
    This function returns the flow graph at time t.

    .. warning::
       Test before use.

    Parameters
    ----------
    W : (N, N) ndarray
        Symmetric adjacency matrix.
    r : (N,) or (N, 1) ndarray, optional
        Rate parameter. Will be set to np.ones((N, 1)) if not specified.
        Default: None
    t : int, optional
        Markov time. Default: 1

    Returns
    -------
    dyn : (N, N) ndarray
        flow graph at time T

    References
    ----------
    .. [1] Lambiotte, R., Sinatra, R., Delvenne, J. C., Evans, T. S.,
       Barahona, M., & Latora, V. (2011). Flow graphs: Interweaving
       dynamics and structure. Physical Review E, 84(1), 017102.
    .. [2] https://github.com/brain-networks/local_scfc/blob/main/fcn/fcn_flow_graph.m
    """
    if r is None:
        r = np.ones((W.shape[0], 1))
    else:
        if r.ndim == 1:
            r = r[:, None]
    deg_wu = np.sum(W, axis=0, keepdims=True)  # (1, N)
    deg_rate = np.sum(deg_wu / r, axis=0, keepdims=True)  # (N, N) => (1, N)
    ps = deg_wu / (deg_rate * r)  # (1, N) / (N, N) => (N, N)
    laplacian = np.diagflat(r) - np.multiply(np.divide(W, deg_wu), r)  # elementwise
    dyn = np.multiply(
        deg_rate * scipy.sparse.linalg.expm(-t * laplacian), ps
    )  # elementwise
    dyn = (dyn + dyn.T) / 2
    return dyn


def _assortativity_und_vectorized(x_vec, weight_mat):
    weight_vec = np.sum(weight_mat, axis=0)
    weight_sum = np.sum(weight_vec)
    x_norm_mean = np.dot(x_vec, weight_vec) / weight_sum
    x_inner = x_vec - x_norm_mean
    upper = np.sum(x_inner * np.squeeze(weight_mat @ x_inner[:, None]))
    lower = np.dot(weight_vec, x_inner**2)
    return upper / lower


def _assortativity_und_numba(x_vec, weight_mat):
    n = len(x_vec)
    weight_vec = np.sum(weight_mat, axis=0)
    weight_sum = np.sum(weight_vec)
    x_norm_mean = np.dot(x_vec, weight_vec) / weight_sum
    upper, lower = 0.0, 0.0
    for i in range(n):
        for j in range(n):
            upper += (
                weight_mat[i, j] * (x_vec[i] - x_norm_mean) * (x_vec[j] - x_norm_mean)
            )
    for i in range(n):
        lower += weight_vec[i] * (x_vec[i] - x_norm_mean) ** 2
    return upper / lower


if has_numba:
    _assortativity_und_numba = njit(_assortativity_und_numba)


def assortativity_und(x, W, use_numba=False):
    r"""
    Calculate assortativity for undirected networks.

    This function implements Bazinet's assortativity for annotated networks as
    defined in [1]_.

    Parameters
    ----------
    x : (N,) array_like
        Annotation values for each node in the network
    W : (N, N) array_like
        Weighted, undirected connection weight array
    use_numba : bool, optional
        Whether to use numba for calculation. Note that numba is not
        necessarily faster than the vectorized (non-numba) version.
        Default: False
        (if numba is available).

    Returns
    -------
    assortativity : float
        Assortativity of the network

    Notes
    -----
    Assortativity is defined as the Pearson correlation between the local
    annotation values of connected nodes [2]_. In other words, it quantifies the
    tendency for nodes with similar annotation values to be connected [1]_.
    For an adjacency matrix :math:`A`, and an annotation vector :math:`\mathbf{x}`,
    the assortativity of a network, with respect to :math:`\bar{\mathbf{x}}`
    is defined as:

    .. math::
        \begin{align}
        r & = \sum_{ij} \frac{a_{ij}}{2m} \tilde{x}_i \tilde{x}_j \\
          & = \sum_{ij} \frac{a_{ij}}{2m}
          (\frac{x_i-\bar{\mathbf{x}}}{\sigma_{\mathbf{x}}})
          (\frac{x_j-\bar{\mathbf{x}}}{\sigma_{\mathbf{x}}})
        \end{align}

    where :math:`a_{ij}` is the weight of the connection between nodes :math:`i`
    and :math:`j`, :math:`2m` is the total weight of the network.
    :math:`\bar{\mathbf{x}}` and :math:`\sigma_{\mathbf{x}}` are the mean and
    standard deviation of the annotation, defined as:

    .. math::
        \begin{align}
        \bar{\mathbf{x}} & = \frac{1}{2m} \sum_i k_i x_i \\
        \sigma_{\mathbf{x}} & = \sqrt{\frac{1}{2m}
        \sum_i k_i (x_i - \bar{\mathbf{x}})^2}
        \end{align}

    in which :math:`k_i` is the sum of the weights of the connections to node :math:`i`.

    References
    ----------
    .. [1] Bazinet, V., Hansen, J. Y., Vos de Wael, R., Bernhardt, B. C., van
        den Heuvel, M. P., & Misic, B. (2023). Assortative mixing in
        micro-architecturally annotated brain connectomes. Nature
        Communications, 14(1), 2850.
    .. [2] Newman, M. E. (2003). Mixing patterns in networks. Physical review E,
        67(2), 026126.

    See Also
    --------
    netneurotools.metrics.assortativity_dir
    """
    if use_numba:
        if not has_numba:
            raise ValueError("Numba not installed; cannot use numba for calculation")
        return _assortativity_und_numba(x, W)
    return _assortativity_und_vectorized(x, W)


def _assortativity_dir_vectorized(x_vec, weight_mat):
    weight_vec_in = np.sum(weight_mat, axis=0)
    weight_vec_out = np.sum(weight_mat, axis=1)
    weight_sum = np.sum(weight_vec_in)
    x_norm_mean_in = np.dot(x_vec, weight_vec_in) / weight_sum
    x_norm_mean_out = np.dot(x_vec, weight_vec_out) / weight_sum
    x_inner_in = x_vec - x_norm_mean_in
    x_inner_out = x_vec - x_norm_mean_out
    upper = np.sum(x_inner_out * np.squeeze(weight_mat @ x_inner_in[:, None]))
    lower_1 = np.dot(weight_vec_in, x_inner_in**2)
    lower_2 = np.dot(weight_vec_out, x_inner_out**2)
    return upper / np.sqrt(lower_1 * lower_2)


def _assortativity_dir_numba(x_vec, weight_mat):
    n = len(x_vec)
    weight_vec_in = np.sum(weight_mat, axis=0)
    weight_vec_out = np.sum(weight_mat, axis=1)
    weight_sum = np.sum(weight_vec_in)
    x_norm_mean_in = np.dot(x_vec, weight_vec_in) / weight_sum
    x_norm_mean_out = np.dot(x_vec, weight_vec_out) / weight_sum
    upper, lower_1, lower_2 = 0.0, 0.0, 0.0
    for i in range(n):
        for j in range(n):
            upper += (
                weight_mat[i, j]
                * (x_vec[j] - x_norm_mean_in)
                * (x_vec[i] - x_norm_mean_out)
            )
    for i in range(n):
        lower_1 += weight_vec_in[i] * (x_vec[i] - x_norm_mean_in) ** 2
        lower_2 += weight_vec_out[i] * (x_vec[i] - x_norm_mean_out) ** 2
    return upper / np.sqrt(lower_1 * lower_2)


if has_numba:
    _assortativity_dir_numba = njit(_assortativity_dir_numba)


def assortativity_dir(x, W, use_numba=False):
    r"""
    Calculate assortativity for directed networks.

    This function implements Bazinet's assortativity for annotated networks as
    defined in [1]_.

    Parameters
    ----------
    x : (N,) array_like
        Annotation values for each node in the network
    W : (N, N) array_like
        Weighted, directed connection weight array
    use_numba : bool, optional
        Whether to use numba for calculation. Note that numba is not
        necessarily faster than the vectorized (non-numba) version.
        Default: False
        (if numba is available).

    Returns
    -------
    assortativity : float
        Assortativity of the network

    Notes
    -----
    For a directed adjacency matrix :math:`A`, and an annotation vector
    :math:`\mathbf{x}`, the assortativity of the network is defined as:

    .. math::

        \begin{align}
        r & = \sum_{ij} \frac{a_{ij}}{m}
        \tilde{x}_i^{\text{out}} \tilde{x}_j^{\text{in}} \\
        & = \sum_{ij} \frac{a_{ij}}{m}
        (\frac{x_i-\bar{\mathbf{x}}^{\text{out}}}{\sigma_{\mathbf{x}}^{\text{out}}})
        (\frac{x_j-\bar{\mathbf{x}}^{\text{in}}}{\sigma_{\mathbf{x}}^{\text{in}}})
        \end{align}

    where :math:`a_{ij}` is the weight of the directed connection from node
    :math:`i` to node :math:`j`, and :math:`m` is the total weight of the
    network. :math:`\bar{\mathbf{x}}^{\text{in}}`,
    :math:`\bar{\mathbf{x}}^{\text{out}}`, :math:`\sigma_{\mathbf{x}}^{\text{in}}`,
    and :math:`\sigma_{\mathbf{x}}^{\text{out}}` are the means and standard deviations
    of the in-degree and out-degree
    weighted annotations, defined as:

    .. math::

        \begin{align}
        \bar{\mathbf{x}}^{\text{in}} & = \frac{1}{m} \sum_i k_i^{\text{in}} x_i \\
        \bar{\mathbf{x}}^{\text{out}} & = \frac{1}{m} \sum_i k_i^{\text{out}} x_i \\
        \sigma_{\mathbf{x}}^{\text{in}} & = \sqrt{\frac{1}{m}
        \sum_i k_i^{\text{in}} (x_i - \bar{\mathbf{x}}^{\text{in}})^2} \\
        \sigma_{\mathbf{x}}^{\text{out}} & = \sqrt{\frac{1}{m}
        \sum_i k_i^{\text{out}} (x_i - \bar{\mathbf{x}}^{\text{out}})^2}
        \end{align}

    in which :math:`k_i^{\text{in}}` is the sum of incoming weights to node
    :math:`i` and :math:`k_i^{\text{out}}` is the sum of outgoing weights from
    node :math:`i`.

    References
    ----------
    .. [1] Bazinet, V., Hansen, J. Y., Vos de Wael, R., Bernhardt, B. C., van
        den Heuvel, M. P., & Misic, B. (2023). Assortative mixing in
        micro-architecturally annotated brain connectomes. Nature
        Communications, 14(1), 2850.
    .. [2] Newman, M. E. (2003). Mixing patterns in networks. Physical review E,
        67(2), 026126.

    See Also
    --------
    netneurotools.metrics.assortativity_und
    """
    if use_numba:
        if not has_numba:
            raise ValueError("Numba not installed; cannot use numba for calculation")
        return _assortativity_dir_numba(x, W)
    return _assortativity_dir_vectorized(x, W)


def matching_ind_und(W):
    """
    Calculate undirected matching index.

    M0 = MATCHING_IND_UND(CIJ) computes matching index for undirected
    graph specified by adjacency matrix CIJ. Matching index is a measure of
    similarity between two nodes' connectivity profiles (excluding their
    mutual connection, should it exist).

    This function is adapted and optimized from the Brain Connectivity Toolbox.

    .. warning::
       Test before use.

    Parameters
    ----------
    W : (N x N) ndarray
        Undirected connection matrix.

    Returns
    -------
    M0 : (N x N) ndarray
        Matching index matrix

    References
    ----------
    .. [1] Goñi, J., Van Den Heuvel, M. P., Avena-Koenigsberger,
       A., Velez de Mendizabal, N., Betzel, R. F., Griffa, A., ... &
       Sporns, O. (2014). Resting-brain functional connectivity predicted
       by analytic measures of network communication. Proceedings of the
       National Academy of Sciences, 111(2), 833-838.
    """
    n = W.shape[0]
    K = np.sum(W, axis=0)
    R_ind = np.nonzero(K)[0]
    N = len(R_ind)
    CIJ = W[np.ix_(R_ind, R_ind)]
    M = np.zeros((N, N))

    for i in range(N):
        c1 = CIJ[i, :]
        use = np.logical_or(c1 > 0, CIJ > 0)
        use[:, i] = False
        np.fill_diagonal(use, False)

        ncon = np.sum((c1 + CIJ) * use, axis=1)
        ncon_and = np.logical_and(np.logical_and(c1 > 0, CIJ > 0), use)
        ncon_and_sum = np.sum(ncon_and, axis=1)
        M[:, i] = 2 * ncon_and_sum / ncon

    np.fill_diagonal(M, 0)
    M[np.isnan(M)] = 0
    M0 = np.zeros((n, n))
    M0[np.ix_(R_ind, R_ind)] = M
    return M0


def rich_feeder_peripheral(x, sc, stat="median"):
    """
    Calculate connectivity values in rich, feeder, and peripheral edges.

    Parameters
    ----------
    x : (N, N) numpy.ndarray
        Symmetric correlation or connectivity matrix
    sc : (N, N) numpy.ndarray
        Binary structural connectivity matrix
    stat : {'mean', 'median'}, optional
        Statistic to use over rich/feeder/peripheral links. Default: 'median'

    Returns
    -------
    rfp : (3, k) numpy.ndarray
        Array of median rich (0), feeder (1), and peripheral (2)
        values, defined by `x`. `k` is the maximum degree defined on `sc`.
    pvals : (3, k) numpy.ndarray
        p-value for each link, computed using Welch's t-test.
        Rich links are compared against non-rich links. Feeder links are
        compared against peripheral links. Peripheral links are compared
        against feeder links. T-test is one-sided.

    Notes
    -----
    This code was written by Justine Hansen who promises to fix and even
    optimize the code should any issues arise, provided you let her know.
    """
    stats = ["mean", "median"]
    if stat not in stats:
        raise ValueError(f"Provided stat {stat} not valid.\
                         Must be one of {stats}")

    nnodes = len(sc)
    mask = np.triu(np.ones(nnodes), 1) > 0
    node_degree = degrees_und(sc)
    k = np.max(node_degree).astype(np.int64)
    rfp_label = np.zeros((len(sc[mask]), k))

    for degthresh in range(k):  # for each degree threshold
        hub_idx = np.where(node_degree >= degthresh)  # find the hubs
        hub = np.zeros([nnodes, 1])
        hub[hub_idx, :] = 1

        rfp = np.zeros([nnodes, nnodes])  # for each link, define rfp
        for edge1 in range(nnodes):
            for edge2 in range(nnodes):
                if hub[edge1] + hub[edge2] == 2:
                    rfp[edge1, edge2] = 1  # rich
                if hub[edge1] + hub[edge2] == 1:
                    rfp[edge1, edge2] = 2  # feeder
                if hub[edge1] + hub[edge2] == 0:
                    rfp[edge1, edge2] = 3  # peripheral
        rfp_label[:, degthresh] = rfp[mask]

    rfp = np.zeros([3, k])
    pvals = np.zeros([3, k])
    for degthresh in range(k):
        redfunc = np.median if stat == "median" else np.mean
        for linktype in range(3):
            rfp[linktype, degthresh] = redfunc(
                x[mask][rfp_label[:, degthresh] == linktype + 1]
            )

        # p-value (one-sided Welch's t-test)
        _, pvals[0, degthresh] = ttest_ind(
            x[mask][rfp_label[:, degthresh] == 1],
            x[mask][rfp_label[:, degthresh] != 1],
            equal_var=False,
            alternative="greater",
        )
        _, pvals[1, degthresh] = ttest_ind(
            x[mask][rfp_label[:, degthresh] == 2],
            x[mask][rfp_label[:, degthresh] == 3],
            equal_var=False,
            alternative="greater",
        )
        _, pvals[2, degthresh] = ttest_ind(
            x[mask][rfp_label[:, degthresh] == 3],
            x[mask][rfp_label[:, degthresh] == 2],
            equal_var=False,
            alternative="greater",
        )

    return rfp, pvals



"""Functions for supporting network metrics."""

import numpy as np
from .. import has_numba
if has_numba:
    from numba import njit


def _fast_binarize(W):
    """
    Binarize a matrix.

    Parameters
    ----------
    W : (N, N) array_like
        Matrix to be binarized

    Returns
    -------
    binarized : (N, N) numpy.ndarray
        Binarized matrix
    """
    return (W > 0) * 1


if has_numba:
    _fast_binarize = njit(_fast_binarize)


def _graph_laplacian(W):
    r"""
    Compute the graph Laplacian of a weighted adjacency matrix.

    Graph Laplacian is defined as the degree matrix minus the adjacency
    matrix :math:`L = D - W`, where :math:`D` is the degree matrix and
    is defined as :math:`D_{ii} = \sum_j W_{ij}`.

    The graph Laplacian matrix :math:`L` has the form of

    .. math::
        L = \begin{bmatrix}
            d_1 & -w_{12} & \cdots & -w_{1n} \\
            -w_{21} & d_2 & \cdots & -w_{2n} \\
            \vdots & \vdots & \ddots & \vdots \\
            -w_{n1} & -w_{n2} & \cdots & d_n
        \end{bmatrix}

    Parameters
    ----------
    W : (N, N) array_like
        Weighted, directed/undirected connection weight/length array

    Returns
    -------
    L : (N, N) numpy.ndarray
        Graph Laplacian of `W`
    """
    D = np.diag(np.sum(W, axis=0))
    return D - W


if has_numba:
    _graph_laplacian = njit(_graph_laplacian)  # ("float64[:,::1](float64[:,::1])")



"""Functions for calculating regression."""

from itertools import combinations

import numpy as np
from tqdm import tqdm
import scipy.stats as sstats
from joblib import Parallel, delayed
from sklearn.linear_model import LinearRegression
from sklearn.utils.validation import check_array


def _add_constant(data):
    """
    Add a constant (i.e., intercept) term to `data`.

    Parameters
    ----------
    data : (N, M) array_like
        Samples by features data array

    Returns
    -------
    data : (N, F) np.ndarray
        Where `F` is `M + 1`

    Examples
    --------
    >>> from netneurotools import stats

    >>> A = np.zeros((5, 5))
    >>> Ac = stats._add_constant(A)
    >>> Ac
    array([[0., 0., 0., 0., 0., 1.],
           [0., 0., 0., 0., 0., 1.],
           [0., 0., 0., 0., 0., 1.],
           [0., 0., 0., 0., 0., 1.],
           [0., 0., 0., 0., 0., 1.]])
    """
    data = check_array(data, ensure_2d=False)
    return np.column_stack([data, np.ones(len(data))])


def residualize(X, Y, Xc=None, Yc=None, normalize=True, add_intercept=True):
    """
    Return residuals of regression equation from `Y ~ X`.

    Parameters
    ----------
    X : (N[, R]) array_like
        Coefficient matrix of `R` variables for `N` subjects
    Y : (N[, F]) array_like
        Dependent variable matrix of `F` variables for `N` subjects
    Xc : (M[, R]) array_like, optional
        Coefficient matrix of `R` variables for `M` subjects. If not specified
        then `X` is used to estimate betas. Default: None
    Yc : (M[, F]) array_like, optional
        Dependent variable matrix of `F` variables for `M` subjects. If not
        specified then `Y` is used to estimate betas. Default: None
    normalize : bool, optional
        Whether to normalize (i.e., z-score) residuals. Will use residuals from
        `Yc ~ Xc` for generating mean and variance. Default: True
    add_intercept : bool, optional
        Whether to add intercept to `X` (and `Xc`, if provided). The intercept
        will not be removed, just used in beta estimation. Default: True

    Returns
    -------
    Yr : (N, F) numpy.ndarray
        Residuals of `Y ~ X`

    Notes
    -----
    If both `Xc` and `Yc` are provided, these are used to calculate betas which
    are then applied to `X` and `Y`.
    """
    if ((Yc is None and Xc is not None) or (Yc is not None and Xc is None)):
        raise ValueError('If processing against a comparative group, you must '
                         'provide both `Xc` and `Yc`.')

    X, Y = np.asarray(X), np.asarray(Y)

    if Yc is None:
        Xc, Yc = X.copy(), Y.copy()
    else:
        Xc, Yc = np.asarray(Xc), np.asarray(Yc)

    # add intercept to regressors if requested and calculate fit
    if add_intercept:
        X, Xc = _add_constant(X), _add_constant(Xc)
    betas, *_ = np.linalg.lstsq(Xc, Yc, rcond=None)

    # remove intercept from regressors and betas for calculation of residuals
    if add_intercept:
        betas = betas[:-1]
        X, Xc = X[:, :-1], Xc[:, :-1]

    # calculate residuals
    Yr = Y - (X @ betas)
    Ycr = Yc - (Xc @ betas)

    if normalize:
        Yr = sstats.zmap(Yr, compare=Ycr)

    return Yr


def get_dominance_stats(X, y, use_adjusted_r_sq=True, verbose=False, n_jobs=1):
    """
    Return the dominance analysis statistics for multilinear regression.

    This is a rewritten & simplified version of [DA1]_. It is briefly
    tested against the original package, but still in early stages.
    Please feel free to report any bugs.

    Warning: Still work-in-progress. Parameters might change!

    Parameters
    ----------
    X : (N, M) array_like
        Input data
    y : (N,) array_like
        Target values
    use_adjusted_r_sq : bool, optional
        Whether to use adjusted r squares. Default: True
    verbose : bool, optional
        Whether to print debug messages. Default: False
    n_jobs : int, optional
        The number of jobs to run in parallel. Default: 1

    Returns
    -------
    model_metrics : dict
        The dominance metrics, currently containing `individual_dominance`,
        `partial_dominance`, `total_dominance`, and `full_r_sq`.
    model_r_sq : dict
        Contains all model r squares

    Notes
    -----
    Example usage

    .. code:: python

        from netneurotools.stats import get_dominance_stats
        from sklearn.datasets import load_boston
        X, y = load_boston(return_X_y=True)
        model_metrics, model_r_sq = get_dominance_stats(X, y)

    To compare with [DA1]_, use `use_adjusted_r_sq=False`

    .. code:: python

        from dominance_analysis import Dominance_Datasets
        from dominance_analysis import Dominance
        boston_dataset=Dominance_Datasets.get_boston()
        dominance_regression=Dominance(data=boston_dataset,
                                       target='House_Price',objective=1)
        incr_variable_rsquare=dominance_regression.incremental_rsquare()
        dominance_regression.dominance_stats()

    References
    ----------
    .. [DA1] https://github.com/dominance-analysis/dominance-analysis

    """
    # this helps to remove one element from a tuple
    def remove_ret(tpl, elem):
        lst = list(tpl)
        lst.remove(elem)
        return tuple(lst)

    # sklearn linear regression wrapper
    def get_reg_r_sq(X, y, use_adjusted_r_sq=True):
        lin_reg = LinearRegression()
        lin_reg.fit(X, y)
        yhat = lin_reg.predict(X)
        SS_Residual = sum((y - yhat) ** 2)
        SS_Total = sum((y - np.mean(y)) ** 2)
        r_squared = 1 - (float(SS_Residual)) / SS_Total
        adjusted_r_squared = 1 - (1 - r_squared) * \
            (len(y) - 1) / (len(y) - X.shape[1] - 1)
        if use_adjusted_r_sq:
            return adjusted_r_squared
        else:
            return r_squared

    # helper function to compute r_sq for a given idx_tuple
    def compute_r_sq(idx_tuple):
        return idx_tuple, get_reg_r_sq(X[:, idx_tuple],
                                       y,
                                       use_adjusted_r_sq=use_adjusted_r_sq)

    # generate all predictor combinations in list (num of predictors) of lists
    n_predictor = X.shape[-1]
    # n_comb_len_group = n_predictor - 1
    predictor_combs = [list(combinations(range(n_predictor), i))
                       for i in range(1, n_predictor + 1)]
    if verbose:
        print(f"[Dominance analysis] Generated \
              {len([v for i in predictor_combs for v in i])} combinations")

    model_r_sq = dict()
    results = Parallel(n_jobs=n_jobs)(
        delayed(compute_r_sq)(idx_tuple)
        for len_group in tqdm(predictor_combs,
                              desc='num-of-predictor loop',
                              disable=not verbose)
        for idx_tuple in tqdm(len_group,
                              desc='insider loop',
                              disable=not verbose))

    # extract r_sq from results
    for idx_tuple, r_sq in results:
        model_r_sq[idx_tuple] = r_sq

    if verbose:
        print(f"[Dominance analysis] Acquired {len(model_r_sq)} r^2's")

    # getting all model metrics
    model_metrics = dict([])

    # individual dominance
    individual_dominance = []
    for i_pred in range(n_predictor):
        individual_dominance.append(model_r_sq[(i_pred,)])
    individual_dominance = np.array(individual_dominance).reshape(1, -1)
    model_metrics["individual_dominance"] = individual_dominance

    # partial dominance
    partial_dominance = [[] for _ in range(n_predictor - 1)]
    for i_len in range(n_predictor - 1):
        i_len_combs = list(combinations(range(n_predictor), i_len + 2))
        for j_node in range(n_predictor):
            j_node_sel = [v for v in i_len_combs if j_node in v]
            reduced_list = [remove_ret(comb, j_node) for comb in j_node_sel]
            diff_values = [
                model_r_sq[j_node_sel[i]] - model_r_sq[reduced_list[i]]
                for i in range(len(reduced_list))]
            partial_dominance[i_len].append(np.mean(diff_values))

    # save partial dominance
    partial_dominance = np.array(partial_dominance)
    model_metrics["partial_dominance"] = partial_dominance
    # get total dominance
    total_dominance = np.mean(
        np.r_[individual_dominance, partial_dominance], axis=0)
    # test and save total dominance
    assert np.allclose(total_dominance.sum(),
                       model_r_sq[tuple(range(n_predictor))]), \
           "Sum of total dominance is not equal to full r square!"
    model_metrics["total_dominance"] = total_dominance
    # save full r^2
    model_metrics["full_r_sq"] = model_r_sq[tuple(range(n_predictor))]

    return model_metrics, model_r_sq



"""For testing netneurotools.metrics.bct functionality."""

import pytest
import numpy as np

from netneurotools import metrics

rs = np.random.RandomState(1234)


def test_communicability_bin():
    """Test communicability_bin function."""
    comm = metrics.communicability_bin(rs.choice([0, 1], size=(100, 100)))
    assert comm.shape == (100, 100)

    with pytest.raises(ValueError):
        metrics.communicability_bin(rs.rand(100, 100))


def test_communicability_wei():
    """Test communicability_wei function."""
    comm = metrics.communicability_wei(rs.rand(100, 100))
    assert comm.shape == (100, 100)
    assert np.allclose(np.diag(comm), 0)



"""Functions for working with CIFTI files."""

from pathlib import Path
import numpy as np
import nibabel as nib

from netneurotools.interface.interface_utils import PARCIGNORE


def describe_cifti(filename):
    """
    Print information about CIFTI file.

    Parameters
    ----------
    filename : str or Path
        Path to CIFTI file.

    Returns
    -------
    None
    """
    if not isinstance(filename, Path):
        filename = Path(filename)

    cifti = nib.load(filename)
    print(f"Filename: {filename.name}", f"Path: {filename}", "", sep="\n")

    header = cifti.header
    n_axes = header.number_of_mapped_indices
    axes_indices = header.mapped_indices
    print(f"Number of axes: {n_axes}, axes indices: {axes_indices}")

    for i_axes in axes_indices:
        curr_axis = header.get_axis(i_axes)
        print(f"Axis {i_axes}: {curr_axis.__class__.__name__}")
        # LabelAxis
        if isinstance(curr_axis, nib.cifti2.LabelAxis):
            for k, v in curr_axis.label[0].items():
                print(f"\t{k}:\t{v[0]}\t\t{v[1]}")
        # BrainModelAxis
        if isinstance(curr_axis, nib.cifti2.BrainModelAxis):
            if curr_axis.nvertices:
                print(f"\tSURFACE: {curr_axis.nvertices}")
                for struct_name, struct_slice, _ in curr_axis.iter_structures():
                    if struct_name in [
                        "CIFTI_STRUCTURE_CORTEX_LEFT",
                        "CIFTI_STRUCTURE_CORTEX_RIGHT",
                    ]:
                        print(f"\t\t{struct_name}\t\t{struct_slice}")
            if curr_axis.volume_shape:
                print(f"\tVOLUME: {curr_axis.volume_shape}")
                for struct_name, struct_slice, _ in curr_axis.iter_structures():
                    if struct_name not in [
                        "CIFTI_STRUCTURE_CORTEX_LEFT",
                        "CIFTI_STRUCTURE_CORTEX_RIGHT",
                    ]:
                        print(f"\t\t{struct_name}\t\t{struct_slice}")


def extract_cifti_volume(data, axis):
    """
    Extract volume data from CIFTI file.

    Adapted from [1]_.

    Parameters
    ----------
    data : np.ndarray
        Data from CIFTI file.
    axis : nibabel.cifti2.BrainModelAxis
        BrainModelAxis object from CIFTI file.

    Returns
    -------
    vol_data : np.ndarray
        Volume data.
    affine : np.ndarray
        Affine matrix.

    References
    ----------
    .. [1] https://nbviewer.org/github/neurohackademy/nh2020-curriculum/blob/master/we-nibabel-markiewicz/NiBabel.ipynb
    """
    assert isinstance(axis, nib.cifti2.BrainModelAxis)

    vol_mask = axis.volume_mask
    vol_shape = axis.volume_shape
    vox_indices = tuple(axis.voxel[vol_mask].T)
    data = data.T[vol_mask]
    vol_data = np.zeros(vol_shape + data.shape[1:], dtype=data.dtype)
    vol_data[vox_indices] = data
    return vol_data, axis.affine


def extract_cifti_surface(data, axis, surf_name):
    """
    Extract surface data from CIFTI file.

    Adapted from [1]_.

    Parameters
    ----------
    data : np.ndarray
        Data from CIFTI file.
    axis : nibabel.cifti2.BrainModelAxis
        BrainModelAxis object from CIFTI file.
    surf_name : str
        Name of surface.

    Returns
    -------
    surf_data : np.ndarray
        Surface data.

    References
    ----------
    .. [1] https://nbviewer.org/github/neurohackademy/nh2020-curriculum/blob/master/we-nibabel-markiewicz/NiBabel.ipynb
    """
    assert isinstance(axis, nib.cifti2.BrainModelAxis)

    found = [
        struct_tuple
        for struct_tuple in axis.iter_structures()
        if struct_tuple[0] == surf_name
    ]
    if len(found) == 0:
        raise ValueError(f"No structure named {surf_name}")
    elif len(found) > 1:
        raise ValueError(
            f"Multiple structures named {surf_name}", f"Structures: {found}"
        )
    else:
        name, data_indices, model = found[0]  # noqa: F841

    data = data.T[data_indices]
    vtx_indices = model.vertex
    surf_data = np.zeros((vtx_indices.max() + 1,) + data.shape[1:], dtype=data.dtype)
    surf_data[vtx_indices] = data

    return surf_data


def extract_cifti_labels(axis, parc_ignore=PARCIGNORE):
    """
    Extract label data from CIFTI file.

    Parameters
    ----------
    axis : nibabel.cifti2.LabelAxis
        LabelAxis object from CIFTI file.
    parc_ignore : list, optional
        List of labels to ignore.

    Returns
    -------
    keys : tuple
        Keys.
    labels : tuple
        Labels.
    """
    assert isinstance(axis, nib.cifti2.LabelAxis)
    keys, labels = list(
        zip(
            *[
                (label_tuple[0], label_tuple[1][0])
                for label_tuple in axis.label[0].items()
                if label_tuple[1][0] not in parc_ignore
            ]
        )
    )
    return keys, labels


def extract_cifti_surface_labels(
        data, label_axis, brainmodel_axis, surf_name, parc_ignore=PARCIGNORE):
    """
    Extract surface data and related labels from CIFTI file.

    Parameters
    ----------
    data : np.ndarray
        Data from CIFTI file.
    label_axis : nibabel.cifti2.LabelAxis
        LabelAxis object from CIFTI file.
    brainmodel_axis : nibabel.cifti2.BrainModelAxis
        BrainModelAxis object from CIFTI file.
    surf_name : str
        Name of surface.
    parc_ignore : list, optional
        List of labels to ignore.

    Returns
    -------
    surf_data : np.ndarray
        Surface data.
    keys : tuple
        Keys.
    labels : tuple
        Labels.
    """
    keys, labels = extract_cifti_labels(label_axis, parc_ignore=parc_ignore)

    surf_data = extract_cifti_surface(data, brainmodel_axis, surf_name)

    keys, labels = zip(
        *[
            (key, label)
            for key, label in zip(keys, labels)
            if key in np.unique(surf_data).astype(int)
        ]
    )

    surf_data = surf_data.squeeze()
    keys = tuple(map(int, keys))
    return surf_data, keys, labels


def deconstruct_cifti(filename, brain_model_axis_index=1):
    """
    Extract volume and surface data from CIFTI file.

    Adapted from [1]_.

    Parameters
    ----------
    filename : str or Path
        Path to CIFTI file.
    brain_model_axis_index : int, optional
        Index of BrainModelAxis in CIFTI file.

    Returns
    -------
    vol_data : np.ndarray
        Volume data.
    surf_left : np.ndarray
        Left hemisphere surface (``CIFTI_STRUCTURE_CORTEX_LEFT``) data.
    surf_right : np.ndarray
        Right hemisphere surface (``CIFTI_STRUCTURE_CORTEX_RIGHT``) data.


    References
    ----------
    .. [1] https://nbviewer.org/github/neurohackademy/nh2020-curriculum/blob/master/we-nibabel-markiewicz/NiBabel.ipynb
    """
    cifti = nib.load(filename)
    data = cifti.get_fdata(dtype=np.float32)
    brain_model_axis = cifti.header.get_axis(brain_model_axis_index)
    if not isinstance(brain_model_axis, nib.cifti2.BrainModelAxis):
        raise ValueError(f"Axis {brain_model_axis_index} is not a BrainModelAxis axis.")
    return (
        extract_cifti_volume(data, brain_model_axis),
        extract_cifti_surface(data, brain_model_axis, "CIFTI_STRUCTURE_CORTEX_LEFT"),
        extract_cifti_surface(data, brain_model_axis, "CIFTI_STRUCTURE_CORTEX_RIGHT"),
    )



"""Functions for supporting network constuction."""

import numpy as np
from scipy.sparse import csgraph


def get_triu(data, k=1):
    """
    Return vectorized version of upper triangle from `data`.

    Parameters
    ----------
    data : (N, N) array_like
        Input data
    k : int, optional
        Which diagonal to select from (where primary diagonal is 0). Default: 1

    Returns
    -------
    triu : (N * N-1 / 2) numpy.ndarray
        Upper triangle of `data`

    Examples
    --------
    >>> from netneurotools import networks

    >>> X = np.array([[1, 0.5, 0.25], [0.5, 1, 0.33], [0.25, 0.33, 1]])
    >>> tri = networks.get_triu(X)
    >>> tri
    array([0.5 , 0.25, 0.33])
    """
    return data[np.triu_indices(len(data), k=k)].copy()


def binarize_network(network, retain=10, keep_diag=False):
    """
    Keep top `retain` % of connections in `network` and binarizes.

    Uses the upper triangle for determining connection percentage, which may
    result in disconnected nodes. If this behavior is not desired see
    :py:func:`netneurotools.networks.threshold_network`.

    Parameters
    ----------
    network : (N, N) array_like
        Input graph
    retain : [0, 100] float, optional
        Percent connections to retain. Default: 10
    keep_diag : bool, optional
        Whether to keep the diagonal instead of setting it to 0. Default: False

    Returns
    -------
    binarized : (N, N) numpy.ndarray
        Binarized, thresholded graph

    See Also
    --------
    netneurotools.networks.threshold_network
    """
    if retain < 0 or retain > 100:
        raise ValueError(
            f'Value provided for `retain` is outside [0, 100]: {retain}'
        )

    prctile = 100 - retain
    triu = get_triu(network)
    thresh = np.percentile(triu, prctile, axis=0, keepdims=True)
    binarized = np.array(network > thresh, dtype=int)

    if not keep_diag:
        binarized[np.diag_indices(len(binarized))] = 0

    return binarized


def threshold_network(network, retain=10):
    """
    Keep top `retain` % of connections in `network` and binarizes.

    Uses a minimum spanning tree to ensure that no nodes are disconnected from
    the resulting thresholded graph

    Parameters
    ----------
    network : (N, N) array_like
        Input graph
    retain : [0, 100] float, optional
        Percent connections to retain. Default: 10

    Returns
    -------
    thresholded : (N, N) numpy.ndarray
        Binarized, thresholded graph

    See Also
    --------
    netneurotools.networks.binarize_network
    """
    if retain < 0 or retain > 100:
        raise ValueError(
            f'Value provided for `retain` must be a percent '
            f'in range [0, 100]. Provided: {retain}'
        )

    # get number of nodes in graph and invert weights (MINIMUM spanning tree)
    nodes = len(network)
    graph = np.triu(network * -1)

    # find MST and count # of edges in graph
    mst = csgraph.minimum_spanning_tree(graph).todense()
    mst_edges = np.sum(mst != 0)

    # determine # of remaining edges and ensure we're not over the limit
    remain = int((retain / 100) * ((nodes * (nodes - 1)) / 2)) - mst_edges
    if remain < 0:
        raise ValueError(
            f'Minimum spanning tree with {mst_edges} edges exceeds desired '
            f'connection density of {retain}% ({remain + mst_edges} edges). Cannot '
            f'proceed with graph creation.'
        )

    # zero out edges already in MST and then get indices of next best edges
    graph -= mst
    inds = get_triu(graph).argsort()[:remain]
    inds = tuple(e[inds] for e in np.triu_indices_from(graph, k=1))

    # add edges to MST, symmetrize, and convert to binary matrix
    mst[inds] = graph[inds]
    mst = np.array((mst + mst.T) != 0, dtype=int)

    return mst



"""Functions for calculating spatial statistics."""

import numpy as np

from .. import has_numba
if has_numba:
    from numba import njit


def _morans_i_vectorized(annot, weight):
    n = annot.shape[0]
    annot_demean = annot - np.mean(annot)
    W = np.sum(weight)
    upper = np.sum(annot_demean * np.squeeze(weight @ annot_demean[:, None]))
    lower = np.sum(annot_demean**2)
    return n * upper / (lower * W)


def _morans_i_numba(annot, weight):
    n = annot.shape[0]
    annot = annot - np.mean(annot)  # all occurances are demean-ed
    upper, lower, W = 0.0, 0.0, 0.0
    for i in range(n):
        lower += annot[i] ** 2
        for j in range(n):
            upper += weight[i, j] * annot[i] * annot[j]
            W += weight[i, j]
    return n * upper / lower / W


if has_numba:
    _morans_i_numba = njit(_morans_i_numba)


def morans_i(annot, weight, use_numba=has_numba):
    r"""
    Calculate Moran's I for spatial autocorrelation.

    Parameters
    ----------
    annot : array-like, shape (n,)
        Array of annotations to calculate Moran's I for.
    weight : array-like, shape (n, n)
        Spatial weight matrix. Note that we do not explicitly check for
        symmetry in the weight matrix, nor zero-diagonal elements.
    use_numba : bool, optional
        Whether to use numba for calculation. Default: True (if numba is
        installed).

    Returns
    -------
    morans_i : float
        Moran's I value for the given annotations and weight matrix.

    Notes
    -----
    Moran's I is calculated as:

    .. math::
        I = \frac{n}{\sum_{i=1}^{n} \sum_{j=1}^{n} w_{ij}} \frac{\sum_{i=1}^{n}
        \sum_{j=1}^{n} w_{ij} (x_i - \bar{x})(x_j - \bar{x})}{\sum_{i=1}^{n}
        (x_i - \bar{x})^2}

    where :math:`n` is the number of observations, :math:`w_{ij}` is the spatial
    weight between observations :math:`i` and :math:`j`, :math:`x_i` is the
    annotation for observation :math:`i`, and :math:`\bar{x}` is the mean
    annotation value.

    The value can be tested using the R pacakge ``spdep``:

    .. code:: R

        x <- rnorm(100)
        m <- matrix(runif(100*100), nrow=100)
        w <- mat2listw(m)
        moran(v, w, 100, Szero(w))
        # or
        moran.test(x, w)

    See Also
    --------
    netneurotools.spatial.spatial_stats.local_morans_i
    """
    if use_numba:
        if not has_numba:
            raise ValueError("Numba not installed; cannot use numba for calculation")
        return _morans_i_numba(annot, weight)
    else:
        return _morans_i_vectorized(annot, weight)


def local_morans_i(annot, weight, use_sampvar=True):
    r"""
    Calculate local Moran's I for spatial autocorrelation.

    Parameters
    ----------
    annot : array-like, shape (n,)
        Array of annotations to calculate Moran's I for.
    weight : array-like, shape (n, n)
        Spatial weight matrix. Note that we do not explicitly check for symmetry
        in the weight matrix, nor zero-diagonal elements.
    use_sampvar : bool, optional
        Whether to use sample variance (n - 1) in calculation. Default: True.

    Returns
    -------
    local_morans_i : array, shape (n,)
        Local Moran's I values for the given annotations and weight matrix.

    Notes
    -----
    Local Moran's I is calculated as:

    .. math::
        I_i = \frac{(x_i - \bar{x})}{\sum_{k=1}^n (x_k - \bar{x})^2/(n-1)}
        \sum_{j=1}^n w_{ij}(x_j - \bar{x})

    where :math:`n` is the number of observations, :math:`w_{ij}` is the spatial
    weight between observations :math:`i` and :math:`j`, :math:`x_i` is the
    annotation for observation :math:`i`, and :math:`\bar{x}` is the mean
    annotation value.

    The value can be tested using the R pacakge ``spdep``:

    .. code:: R

            x <- rnorm(100)
            m <- matrix(runif(100*100), nrow=100)
            localmoran(x, mat2listw(m), mlvar=TRUE)

    See Also
    --------
    netneurotools.spatial.spatial_stats.morans_i
    """
    n = annot.shape[0]
    annot_demean = annot - np.mean(annot)
    if use_sampvar:
        lower = n - 1
    else:
        lower = n
    m_2 = np.dot(annot_demean, annot_demean) / lower
    right = np.squeeze(weight @ annot_demean[:, None])  # np.dot(weight, annot)
    return annot_demean * right / m_2


def _gearys_c_vectorized(annot, weight):
    n = annot.shape[0]
    annot_demean = annot - np.mean(annot)
    W = np.sum(weight)
    upper = np.sum(weight * (annot[:, np.newaxis] - annot[np.newaxis, :]) ** 2)
    lower = np.sum(annot_demean**2)
    return (n - 1) * upper / (2 * W * lower)


def _gearys_c_numba(annot, weight):
    n = annot.shape[0]
    annot_demean = annot - np.mean(annot)  # not all occurances are demean-ed
    upper, lower, W = 0.0, 0.0, 0.0
    for i in range(n):
        lower += annot_demean[i] ** 2
        for j in range(n):
            upper += weight[i, j] * (annot[i] - annot[j]) ** 2
            W += weight[i, j]
    return (n - 1) * upper / lower / (2 * W)


if has_numba:
    _gearys_c_numba = njit(_gearys_c_numba)


def gearys_c(annot, weight, use_numba=has_numba):
    r"""
    Calculate Geary's C for spatial autocorrelation.

    Parameters
    ----------
    annot : array-like, shape (n,)
        Array of annotations to calculate Geary's C for.
    weight : array-like, shape (n, n)
        Spatial weight matrix. Note that we do not explicitly check for symmetry
        in the weight matrix, nor zero-diagonal elements.
    use_numba : bool, optional
        Whether to use numba for calculation. Default: True (if numba is
        installed).

    Returns
    -------
    gearys_c : float
        Geary's C value for the given annotations and weight matrix.

    Notes
    -----
    Geary's C is calculated as:

    .. math::
        C = \frac{(n-1)}{2\sum_{i=1}^n \sum_{j=1}^n w_{ij}}
        \frac{\sum_{i=1}^n \sum_{j=1}^n w_{ij}(x_i - x_j)^2}
        {\sum_{i=1}^n(x_i - \bar{x})^2}

    where :math:`n` is the number of observations, :math:`w_{ij}` is the spatial
    weight between observations :math:`i` and :math:`j`, :math:`x_i` is the
    annotation for observation :math:`i`, and :math:`\bar{x}` is the mean
    annotation value.

    The value can be tested using the R pacakge ``spdep``:

    .. code:: R

        x <- rnorm(100)
        m <- matrix(runif(100*100), nrow=100)
        w <- mat2listw(m)
        geary(x, w, 100, 100-1, Szero(w))
        # or
        geary.test(x, w)

    """
    if use_numba:
        if not has_numba:
            raise ValueError("Numba not installed; cannot use numba for calculation")
        return _gearys_c_numba(annot, weight)
    else:
        return _gearys_c_vectorized(annot, weight)


def local_gearys_c(annot, weight, use_sampvar=True):
    r"""
    Calculate local Geary's C for spatial autocorrelation.

    Parameters
    ----------
    annot : array-like, shape (n,)
        Array of annotations to calculate Geary's C for.
    weight : array-like, shape (n, n)
        Spatial weight matrix. Note that we do not explicitly check for symmetry
        in the weight matrix, nor zero-diagonal elements.
    use_sampvar : bool, optional
        Whether to use sample variance (n - 1) in calculation. Default: True.

    Returns
    -------
    local_gearys_c : array, shape (n,)
        Local Geary's C values for the given annotations and weight matrix.

    Notes
    -----
    Local Geary's C is calculated as:

    .. math::
        C_i = \frac{w_{ij} (x_i - x_j)^2}{\sum_{j=1}^n w_{ij} (x_i - x_j)^2}

    where :math:`n` is the number of observations, :math:`w_{ij}` is the spatial
    weight between observations :math:`i` and :math:`j`, :math:`x_i` is the
    annotation for observation :math:`i`, and :math:`\bar{x}` is the mean
    annotation value.

    The value can be tested using the R pacakge ``spdep``:

    .. code:: R

        x <- rnorm(100)
        m <- matrix(runif(100*100), nrow=100)
        localC(v, mat2listw(m))
    """
    n = annot.shape[0]
    annot_demean = annot - np.mean(annot)
    if use_sampvar:
        lower = n - 1
    else:
        lower = n
    m_2 = np.dot(annot_demean, annot_demean) / lower  # unbiased
    annot_diffsq = (annot[:, None] - annot[None, :]) ** 2
    right = np.sum(weight * annot_diffsq, axis=1)
    return right / m_2


def _lees_l_vectorized(annot_1, annot_2, weight):
    n = annot_1.shape[0]
    annot_1_demean = annot_1 - np.mean(annot_1)
    annot_2_demean = annot_2 - np.mean(annot_2)

    # Calculate upper term
    w_x1 = weight @ annot_1_demean
    w_x2 = weight @ annot_2_demean
    upper = np.sum(w_x1 * w_x2)

    # Calculate S2 (sum of squared row sums)
    s2 = np.sum(np.sum(weight, axis=1) ** 2)

    # Calculate lower terms
    lower_1 = np.sum(annot_1_demean ** 2)
    lower_2 = np.sum(annot_2_demean ** 2)

    return n / s2 * upper / np.sqrt(lower_1) / np.sqrt(lower_2)


def _lees_l_numba(annot_1, annot_2, weight):
    n = annot_1.shape[0]
    annot_1_demean = annot_1 - np.mean(annot_1)
    annot_2_demean = annot_2 - np.mean(annot_2)

    upper, lower_1, lower_2, s2 = 0.0, 0.0, 0.0, 0.0
    for i in range(n):
        upper_1, upper_2 = 0.0, 0.0
        s2_tmp = 0.0
        for j in range(n):
            upper_1 += weight[i, j] * annot_1_demean[j]
            upper_2 += weight[i, j] * annot_2_demean[j]
            s2_tmp += weight[i, j]
        upper += upper_1 * upper_2
        s2 += s2_tmp ** 2
        #
        lower_1 += annot_1_demean[i] ** 2
        lower_2 += annot_2_demean[i] ** 2

    return n / s2 * upper / np.sqrt(lower_1) / np.sqrt(lower_2)


def lees_l(annot_1, annot_2, weight, use_numba=has_numba):
    r"""
    Calculate Lee's L for spatial autocorrelation.

    Parameters
    ----------
    annot_1 : array-like, shape (n,)
        Array of annotations to calculate Lee's L for.
    annot_2 : array-like, shape (n,)
        Array of annotations to calculate Lee's L for.
    weight : array-like, shape (n, n)
        Spatial weight matrix. Note that we do not explicitly check for symmetry
        in the weight matrix, nor zero-diagonal elements.
    use_numba : bool, optional
        Whether to use numba for calculation. Default: True (if numba is
        installed).

    Returns
    -------
    lees_l : float
        Lee's L value for the given annotations and weight matrix.

    Notes
    -----
    Lee's L is calculated as:

    .. math::
        L(x,y) = \frac{n}{\sum_{i=1}^n(\sum_{j=1}^n w_{ij})^2}
        \frac{\sum_{i=1}^n(\sum_{j=1}^n w_{ij}(x_i - \bar{x}))
        (\sum_{j=1}^n w_{ij}(y_i - \bar{y}))}
        {\sqrt{\sum_{i=1}^n(x_i - \bar{x})^2}\sqrt{\sum_{i=1}^n(y_i - \bar{y})^2}}

    where :math:`n` is the number of observations, :math:`w_{ij}` is the spatial
    weight between observations :math:`i` and :math:`j`, :math:`x_i` is the
    annotation for observation :math:`i`, and :math:`\bar{x}` is the mean annotation
    value. :math:`x` and :math:`y` are the annotations for the two variables.

    The value can be tested using the R pacakge ``spdep``:

    .. code:: R

        x <- rnorm(100)
        y <- rnorm(100)
        m <- matrix(runif(100*100), nrow=100)
        lee(x, y, mat2listw(m), 100)
        # or
        lee.test(x, y, mat2listw(m))
    """
    if use_numba:
        if not has_numba:
            raise ValueError("Numba not installed; cannot use numba for calculation")
        return _lees_l_numba(annot_1, annot_2, weight)
    else:
        return _lees_l_vectorized(annot_1, annot_2, weight)


def local_lees_l(annot_1, annot_2, weight):
    r"""
    Calculate local Lee's L for spatial autocorrelation.

    Parameters
    ----------
    annot_1 : array-like, shape (n,)
        Array of annotations to calculate Lee's L for.
    annot_2 : array-like, shape (n,)
        Array of annotations to calculate Lee's L for.
    weight : array-like, shape (n, n)
        Spatial weight matrix. Note that we do not explicitly check for symmetry
        in the weight matrix, nor zero-diagonal elements.

    Returns
    -------
    local_lees_l : array, shape (n,)
        Local Lee's L values for the given annotations and weight matrix.

    Notes
    -----
    Local Lee's L is calculated as:

    .. math::
        L_i(x,y) = \frac{(\sum_{j=1}^n w_{ij}(x_i - \bar{x}))
        (\sum_{j=1}^n w_{ij}(y_i - \bar{y}))}
        {\sqrt{\sum_{i=1}^n(x_i - \bar{x})^2}\sqrt{\sum_{i=1}^n(y_i - \bar{y})^2}}

    where :math:`n` is the number of observations, :math:`w_{ij}` is the spatial
    weight between observations :math:`i` and :math:`j`, :math:`x_i` is the
    annotation for observation :math:`i`, and :math:`\bar{x}` is the mean annotation
    value. :math:`x` and :math:`y` are the annotations for the two variables.

    The value can be tested using the R pacakge ``spdep``:

    .. code:: R

        x <- rnorm(100)
        y <- rnorm(100)
        m <- matrix(runif(100*100), nrow=100)
        lee(x, y, mat2listw(m), 100)
    """
    n = annot_1.shape[0]
    annot_1_demean = annot_1 - np.mean(annot_1)
    annot_2_demean = annot_2 - np.mean(annot_2)
    lower_1 = np.sqrt(np.sum(annot_1_demean ** 2))
    lower_2 = np.sqrt(np.sum(annot_2_demean ** 2))
    upper_1 = np.squeeze(weight @ annot_1_demean[:, None])  # np.dot(weight, annot)
    upper_2 = np.squeeze(weight @ annot_2_demean[:, None])  # np.dot(weight, annot)
    return n * upper_1 * upper_2 / lower_1 / lower_2



"""Functional tests for the transforms module."""

import numpy as np
import pytest
from netneurotools.interface.surf_parc import (
    parcels_to_vertices,  # noqa: F401
    vertices_to_parcels,
)
from netneurotools.datasets import fetch_schaefer2018, fetch_cammoun2012

_parc_to_nvert = {
    "annot": 10242,
    "dlabel.nii": 32492,
    "gii": 32492,
}
_parc_to_nparc = {
    "annot": 50,
    "dlabel.nii": 50,
    "gii": 34,
}


test_params, test_ids = zip(
    *[
        (
            (
                {"file_type": file_type, "hemi_type": hemi_type},
                {
                    "file_type": file_type,
                    "hemi_type": hemi_type,
                    "data_width": data_width,
                },
            ),
            f"{file_type}-{hemi_type}-{data_width}d",
        )
        for file_type in ["annot", "dlabel.nii", "gii"]
        for hemi_type in ["both", "L", "R"]
        for data_width in [0, 1, 2]
    ]
)


@pytest.fixture(scope="session")
def prepare_parc_files(request, tmp_path_factory):
    """Prepare parcellation files for parcels_to_vertices tests."""
    file_type = request.param["file_type"]
    hemi_type = request.param["hemi_type"]
    data_dir = tmp_path_factory.mktemp("data")

    if file_type == "annot":
        parc = fetch_schaefer2018(version="fsaverage5", data_dir=data_dir, verbose=0)[
            "100Parcels7Networks"
        ]
    elif file_type == "dlabel.nii":
        parc = fetch_schaefer2018(version="fslr32k", data_dir=data_dir, verbose=0)[
            "100Parcels7Networks"
        ]
    elif file_type == "gii":
        parc = fetch_cammoun2012(version="fslr32k", data_dir=data_dir, verbose=0)[
            "scale033"
        ]
    else:
        raise ValueError(f"Invalid file type: {file_type}")

    if file_type == "dlabel.nii":
        parc_file = parc
    else:
        if hemi_type == "both":
            parc_file = (parc.L, parc.R)
        elif hemi_type == "L":
            parc_file = parc.L
        elif hemi_type == "R":
            parc_file = parc.R
        else:
            raise ValueError(f"Invalid hemisphere type: {hemi_type}")

    return parc_file


def _generate_random_data(which, file_type, hemi_type, data_width):
    if which == "vertices_to_parcels":
        n = _parc_to_nvert[file_type]
    elif which == "parcels_to_vertices":
        n = _parc_to_nparc[file_type]
    else:
        raise ValueError(f"Invalid function: {which}")

    rng = np.random.default_rng(sum(map(ord, file_type + hemi_type)))
    data = (rng.random(n), rng.random(n))

    if data_width == 0:
        pass
    elif data_width == 1:
        data = (data[0][:, np.newaxis], data[1][:, np.newaxis])
    elif data_width == 2:
        data = (np.c_[data[0], data[0]], np.c_[data[1], data[1]])
    else:
        raise ValueError(f"Invalid data dimension: {data_width}")

    if hemi_type == "both":
        data = data
    elif hemi_type == "L":
        data = data[0]
    elif hemi_type == "R":
        data = data[1]
    else:
        raise ValueError(f"Invalid hemisphere type: {hemi_type}")

    return data


@pytest.fixture(scope="session")
def generate_vertices_to_parcels_data(request):
    """Generate data for vertices_to_parcels tests."""
    file_type = request.param["file_type"]
    hemi_type = request.param["hemi_type"]
    data_width = request.param["data_width"]

    vert_data = _generate_random_data(
        "vertices_to_parcels", file_type, hemi_type, data_width
    )

    expected_all = {
        "annot": {
            "both": (100, 0.4989233),
            "L": (50, 0.49557236),
            "R": (50, 0.49596187),
        },
        "dlabel.nii": {
            "both": (100, 0.4977689),
            "L": (50, 0.49637252),
            "R": (50, 0.50187889),
        },
        "gii": {
            "both": (68, 0.4966058),
            "L": (34, 0.49639294),
            "R": (34, 0.5032322),
        },
    }
    # generate expected output
    expected = expected_all[file_type][hemi_type]

    return vert_data, expected


@pytest.mark.parametrize(
    "prepare_parc_files, generate_vertices_to_parcels_data",
    test_params,
    ids=test_ids,
    indirect=True,
)
def test_vertices_to_parcels(
    prepare_parc_files, generate_vertices_to_parcels_data, request
):
    """Test vertices_to_parcels function."""
    # curr_id = request.node.callspec.id
    curr_params = request.node.callspec.params["generate_vertices_to_parcels_data"]
    # file_type = curr_params["file_type"]
    hemi_type = curr_params["hemi_type"]
    data_width = curr_params["data_width"]

    parc_file = prepare_parc_files
    vert_data, expected = generate_vertices_to_parcels_data

    reduced, keys, labels = vertices_to_parcels(
        vert_data, parc_file, hemi=hemi_type, background=0
    )
    if hemi_type == "both":
        # test output shape and type
        assert len(reduced) == 2
        assert all([isinstance(k, np.ndarray) for k in reduced])
        assert len(keys) == 2
        assert all([isinstance(k, tuple) for k in keys])
        assert len(labels) == 2
        assert all([isinstance(k, tuple) for k in labels])
        # test output data
        if data_width == 0:
            assert all([r.ndim == 1 for r in reduced])
        else:
            assert all([r.ndim == 2 for r in reduced])
            assert all([r.shape[1] == data_width for r in reduced])
        assert sum([r.shape[0] for r in reduced]) == expected[0]
        assert np.allclose(np.mean(np.concatenate(reduced), axis=0), expected[1])
    else:
        # test output shape and type
        assert isinstance(reduced, np.ndarray)
        assert isinstance(keys, tuple)
        assert isinstance(labels, tuple)
        # test output data
        if data_width == 0:
            assert reduced.ndim == 1
        else:
            assert reduced.ndim == 2
            assert reduced.shape[1] == data_width

        assert reduced.shape[0] == expected[0]
        assert np.allclose(np.mean(reduced, axis=0), expected[1])


@pytest.fixture(scope="session")
def generate_parcels_to_vertices_data(request):
    """Generate data for parcels_to_vertices tests."""
    file_type = request.param["file_type"]
    hemi_type = request.param["hemi_type"]
    data_width = request.param["data_width"]

    parcels_data = _generate_random_data(
        "parcels_to_vertices", file_type, hemi_type, data_width
    )

    expected_all = {
        "annot": {
            "both": (20484, 0.48047245),
            "L": (10242, 0.5117257),
            "R": (10242, 0.49651137),
        },
        "dlabel.nii": {
            "both": (64984, 0.52391384),
            "L": (32492, 0.51929605),
            "R": (32492, 0.4648587),
        },
        "gii": {
            "both": (64984, 0.5227571),
            "L": (32492, 0.54153746),
            "R": (32492, 0.5344661),
        },
    }
    # generate expected output
    expected = expected_all[file_type][hemi_type]

    return parcels_data, expected


@pytest.mark.parametrize(
    "prepare_parc_files, generate_parcels_to_vertices_data",
    test_params,
    ids=test_ids,
    indirect=True,
)
def test_parcels_to_vertices(
    prepare_parc_files, generate_parcels_to_vertices_data, request
):
    """Test parcels_to_vertices function."""
    curr_params = request.node.callspec.params["generate_parcels_to_vertices_data"]
    hemi_type = curr_params["hemi_type"]
    data_width = curr_params["data_width"]

    parc_file = prepare_parc_files
    parc_data, expected = generate_parcels_to_vertices_data

    projected, keys, labels = parcels_to_vertices(
        parc_data, parc_file, hemi=hemi_type, fill=np.nan
    )
    if hemi_type == "both":
        # test output shape and type
        assert len(projected) == 2
        assert all([isinstance(k, np.ndarray) for k in projected])
        assert len(keys) == 2
        assert all([isinstance(k, tuple) for k in keys])
        assert len(labels) == 2
        assert all([isinstance(k, tuple) for k in labels])
        # test output data
        if data_width == 0:
            assert all([r.ndim == 1 for r in projected])
        else:
            assert all([r.ndim == 2 for r in projected])
            assert all([r.shape[1] == data_width for r in projected])
        assert sum([r.shape[0] for r in projected]) == expected[0]
        assert np.allclose(np.nanmean(np.concatenate(projected), axis=0), expected[1])
    else:
        # test output shape and type
        assert isinstance(projected, np.ndarray)
        assert isinstance(keys, tuple)
        assert isinstance(labels, tuple)
        # test output data
        if data_width == 0:
            assert projected.ndim == 1
        else:
            assert projected.ndim == 2
            assert projected.shape[1] == data_width

        assert projected.shape[0] == expected[0]
        assert np.allclose(np.nanmean(projected, axis=0), expected[1])



"""Functions for matplotlib-based plotting."""

from typing import Iterable
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches


def _grid_communities(communities):
    """
    Generate boundaries of `communities`.

    Parameters
    ----------
    communities : array_like
        Community assignment vector

    Returns
    -------
    bounds : list
        Boundaries of communities
    """
    communities = np.asarray(communities)
    if 0 in communities:
        communities = communities + 1

    comm = communities[np.argsort(communities)]
    bounds = []
    for i in np.unique(comm):
        ind = np.where(comm == i)
        if len(ind) > 0:
            bounds.append(np.min(ind))

    bounds.append(len(communities))

    return bounds


def _sort_communities(consensus, communities):
    """
    Sort `communities` in `consensus` according to strength.

    Parameters
    ----------
    consensus : array_like
        Correlation matrix
    communities : array_like
        Community assignments for `consensus`

    Returns
    -------
    inds : np.ndarray
        Index array for sorting `consensus`
    """
    communities = np.asarray(communities)
    if 0 in communities:
        communities = communities + 1

    bounds = _grid_communities(communities)
    inds = np.argsort(communities)

    for n, f in enumerate(bounds[:-1]):
        i = inds[f:bounds[n + 1]]
        cco = i[consensus[np.ix_(i, i)].mean(axis=1).argsort()[::-1]]
        inds[f:bounds[n + 1]] = cco

    return inds


def plot_mod_heatmap(data, communities, *, inds=None, edgecolor='black',
                     ax=None, figsize=(6.4, 4.8), xlabels=None, ylabels=None,
                     xlabelrotation=90, ylabelrotation=0, cbar=True,
                     square=True, xticklabels=None, yticklabels=None,
                     mask_diagonal=True, **kwargs):
    """
    Plot `data` as heatmap with borders drawn around `communities`.

    Parameters
    ----------
    data : (N, N) array_like
        Correlation matrix
    communities : (N,) array_like
        Community assignments for `data`
    inds : (N,) array_like, optional
        Index array for sorting `data` within `communities`. If None, these
        will be generated from `data`. Default: None
    edgecolor : str, optional
        Color for lines demarcating community boundaries. Default: 'black'
    ax : matplotlib.axes.Axes, optional
        Axis on which to plot the heatmap. If none provided, a new figure and
        axis will be created. Default: None
    figsize : tuple, optional
        Size of figure to create if `ax` is not provided. Default: (20, 20)
    {x,y}labels : list, optional
        List of labels on {x,y}-axis for each community in `communities`. The
        number of labels should match the number of unique communities.
        Default: None
    {x,y}labelrotation : float, optional
        Angle of the rotation of the labels. Available only if `{x,y}labels`
        provided. Default : xlabelrotation: 90, ylabelrotation: 0
    square : bool, optional
        Setting the matrix with equal aspect. Default: True
    {x,y}ticklabels : list, optional
        Incompatible with `{x,y}labels`. List of labels for each entry (not
        community) in `data`. Default: None
    cbar : bool, optional
        Whether to plot colorbar. Default: True
    mask_diagonal : bool, optional
        Whether to mask the diagonal in the plotted heatmap. Default: True
    kwargs : key-value mapping
        Keyword arguments for `plt.pcolormesh()`

    Returns
    -------
    ax : matplotlib.axes.Axes
        Axis object containing plot
    """
    for t, label in zip([xticklabels, yticklabels], [xlabels, ylabels]):
        if t is not None and label is not None:
            raise ValueError('Cannot set both {x,y}labels and {x,y}ticklabels')

    # get indices for sorting consensus
    if inds is None:
        inds = _sort_communities(data, communities)

    if ax is None:
        _, ax = plt.subplots(1, 1, figsize=figsize)

    # plot data re-ordered based on community and node strength
    if mask_diagonal:
        plot_data = np.ma.masked_where(np.eye(len(data)),
                                       data[np.ix_(inds, inds)])
    else:
        plot_data = data[np.ix_(inds, inds)]

    coll = ax.pcolormesh(plot_data, edgecolor='none', **kwargs)
    ax.set(xlim=(0, plot_data.shape[1]), ylim=(0, plot_data.shape[0]))

    # set equal aspect
    if square:
        ax.set_aspect('equal')

    for side in ['top', 'right', 'left', 'bottom']:
        ax.spines[side].set_visible(False)

    # invert the y-axis so it looks "as expected"
    ax.invert_yaxis()

    # plot the colorbar
    if cbar:
        cb = ax.figure.colorbar(coll)
        if kwargs.get('rasterized', False):
            cb.solids.set_rasterized(True)

    # draw borders around communities
    bounds = _grid_communities(communities)
    bounds[0] += 0.2
    bounds[-1] -= 0.2
    for n, edge in enumerate(np.diff(bounds)):
        ax.add_patch(mpatches.Rectangle((bounds[n], bounds[n]),
                                       edge, edge, fill=False, linewidth=2,
                                       edgecolor=edgecolor))

    if xlabels is not None or ylabels is not None:
        # find the tick locations
        initloc = _grid_communities(communities)
        tickloc = []
        for loc in range(len(initloc) - 1):
            tickloc.append(np.mean((initloc[loc], initloc[loc + 1])))

        if xlabels is not None:
            # make sure number of labels match the number of ticks
            if len(tickloc) != len(xlabels):
                raise ValueError('Number of labels do not match the number of '
                                 'unique communities.')
            else:
                ax.set_xticks(tickloc)
                ax.set_xticklabels(labels=xlabels, rotation=xlabelrotation)
                ax.tick_params(left=False, bottom=False)
        if ylabels is not None:
            # make sure number of labels match the number of ticks
            if len(tickloc) != len(ylabels):
                raise ValueError('Number of labels do not match the number of '
                                 'unique communities.')
            else:
                ax.set_yticks(tickloc)
                ax.set_yticklabels(labels=ylabels, rotation=ylabelrotation)
                ax.tick_params(left=False, bottom=False)

    if xticklabels is not None:
        labels_ind = [xticklabels[i] for i in inds]
        ax.set_xticks(np.arange(len(labels_ind)) + 0.5)
        ax.set_xticklabels(labels_ind, rotation=90)
    if yticklabels is not None:
        labels_ind = [yticklabels[i] for i in inds]
        ax.set_yticks(np.arange(len(labels_ind)) + 0.5)
        ax.set_yticklabels(labels_ind)

    return ax


def plot_point_brain(data, coords, views=None, views_orientation='vertical',
                     views_size=(4, 2.4), cbar=False, robust=True, size=50,
                     **kwargs):
    """
    Plot `data` as a cloud of points in 3D space based on specified `coords`.

    Parameters
    ----------
    data : (N,) array_like
        Data for an `N` node parcellation; determines color of points
    coords : (N, 3) array_like
        x, y, z coordinates for `N` node parcellation
    views : list, optional
        List specifying which views to use. Can be any of {'sagittal', 'sag',
        'coronal', 'cor', 'axial', 'ax'}. If not specified will use 'sagittal'
        and 'axial'. Default: None
    views_orientation: str, optional
        Orientation of the views. Can be either 'vertical' or 'horizontal'.
        Default: 'vertical'.
    views_size : tuple, optional
        Figure size of each view. Default: (4, 2.4)
    cbar : bool, optional
        Whether to also show colorbar. Default: False
    robust : bool, optional
        Whether to use robust calculation of `vmin` and `vmax` for color scale.
    size : int, optional
        Size of points on plot. Default: 50
    **kwargs
        Key-value pairs passed to `matplotlib.axes.Axis.scatter`

    Returns
    -------
    fig : :class:`matplotlib.figure.Figure`
        Figure object for the plot
    axes : :class:`matplotlib.axes.Axes`
        Axes object for the plot
    """
    _views = dict(sagittal=(0, 180), sag=(0, 180),
                  axial=(90, 180), ax=(90, 180),
                  coronal=(0, 90), cor=(0, 90))

    x, y, z = coords[:, 0], coords[:, 1], coords[:, 2]

    if views is None:
        views = [_views[f] for f in ['sagittal', 'axial']]
    else:
        if not isinstance(views, Iterable) or isinstance(views, str):
            views = [views]
        views = [_views[f] for f in views]

    if views_orientation == 'vertical':
        ncols, nrows = 1, len(views)
    elif views_orientation == 'horizontal':
        ncols, nrows = len(views), 1
    figsize = (ncols * views_size[0], nrows * views_size[1])

    # create figure and axes (3d projections)
    fig, axes = plt.subplots(ncols=ncols, nrows=nrows,
                             figsize=figsize,
                             subplot_kw=dict(projection='3d'))

    opts = dict(linewidth=0.5, edgecolor='gray', cmap='viridis')
    if robust:
        vmin, vmax = np.percentile(data, [2.5, 97.5])
        opts.update(dict(vmin=vmin, vmax=vmax))
    opts.update(kwargs)

    # iterate through saggital/axial views and plot, rotating as needed
    for n, view in enumerate(views):
        # if only one view then axes is not a list!
        ax = axes[n] if len(views) > 1 else axes
        # make the actual scatterplot and update the view / aspect ratios
        col = ax.scatter(x, y, z, c=data, s=size, **opts)
        ax.view_init(*view)
        ax.axis('off')
        scaling = np.array([ax.get_xlim(),
                            ax.get_ylim(),
                            ax.get_zlim()])
        ax.set_box_aspect(tuple(scaling[:, 1] - scaling[:, 0]))

    fig.subplots_adjust(left=0, right=1, bottom=0, top=1, hspace=0, wspace=0)

    # add colorbar to axes
    if cbar:
        cbar = fig.colorbar(col, ax=axes.flatten(),
                            drawedges=False, shrink=0.7)
        cbar.outline.set_linewidth(0)

    return fig, axes


def plot_simple_brain():
    """Plot a simple brain using matplotlib."""
    # https://github.com/dutchconnectomelab/Simple-Brain-Plot
    pass



"""For testing netneurotools.modularity.modules functionality."""

import bct
import pytest
import numpy as np
from sklearn.cluster import k_means, spectral_clustering

from netneurotools import modularity

rs = np.random.RandomState(1234)


@pytest.mark.parametrize('c1, c2, out', [
    # uniform communities
    (np.array([0, 0, 0, 0, 0, 0, 0, 0, 0]),
     np.array([1, 1, 1, 1, 1, 1, 1, 1, 1]),
     np.array([1, 1, 1, 1, 1, 1, 1, 1, 1])),
    # perfectly matched but mislabeled communities
    (np.array([1, 1, 1, 2, 2, 2, 3, 3, 3]),
     np.array([2, 2, 2, 1, 1, 1, 3, 3, 3]),
     np.array([2, 2, 2, 1, 1, 1, 3, 3, 3])),
    # 1 cluster --> 2 clusters
    (np.array([1, 1, 1, 1, 1, 1, 1, 1, 1]),
     np.array([1, 1, 1, 1, 2, 2, 2, 2, 2]),
     np.array([2, 2, 2, 2, 2, 2, 2, 2, 2])),
    # 3 clusters --> 2 clusters
    (np.array([1, 1, 1, 2, 2, 2, 3, 3, 3]),
     np.array([1, 1, 1, 1, 1, 2, 2, 2, 2]),
     np.array([1, 1, 1, 3, 3, 3, 2, 2, 2]))
])
def test_match_cluster_labels(c1, c2, out):
    """Test matching of cluster labels."""
    assert np.all(modularity.match_cluster_labels(c1, c2) == out)


def test_match_assignments():
    """Test matching of clustering assignments."""
    # generate some random data to be clustered (must be symmetric)
    rs = np.random.RandomState(1234)
    data = rs.rand(100, 100)
    data = (data + data.T) / 2

    # cluster it 200 times (spectral clustering provides unstable labels but
    # should provide ~stable(ish) assignments)
    assignments = [spectral_clustering(data, n_clusters=2, random_state=rs)
                   for n in range(200)]
    assignments = np.column_stack(assignments)

    # make sure the labels are, in fact, unstable
    assert not np.all(assignments[:, [0]] == assignments)

    # match labels and assert that we got perfect matches (this is not 100%
    # guaranteed with spectral clustering but it is...pretty likely)
    matched = modularity.match_assignments(assignments, seed=rs)
    assert np.all(matched[:, [0]] == matched)

    # check that we didn't _actually_ change cluster assignments with matching;
    # the agreement matrices should match!
    assert np.allclose(bct.clustering.agreement(assignments),
                       bct.clustering.agreement(matched))


def test_reorder_assignments():
    """Test re-ordering of clustering assignments."""
    # generate a bunch of ~random(ish) clustering assignments that have a bit
    # of consistency but aren't all identical
    rs = np.random.RandomState(1234)
    data = rs.rand(100, 200)
    assignments = [k_means(data, n_clusters=2)[1] for n in range(200)]
    assignments = np.column_stack(assignments)

    # make sure the labels are quite unstable
    assert not np.all(assignments[:, [0]] == assignments)

    # re-order labels and assert that we still _don't_ have perfect matches
    # (we're re-labelling the matrix but k-means does not provide stable
    # clustering assignments so we shouldn't get identical assignments even
    # after "matching")
    reordered, idx = modularity.reorder_assignments(assignments, seed=1234)
    assert not np.all(reordered[:, [0]] == reordered)

    # make sure that the returned idx does exactly what it's supposed to
    matched = modularity.match_assignments(assignments, seed=1234)[idx]
    assert np.all(matched == reordered)


@pytest.mark.parametrize('assignments, clusters', [
    (np.array([[0, 1, 2], [0, 1, 2], [0, 1, 2],
               [1, 2, 0], [1, 2, 0], [1, 2, 0],
               [2, 0, 1], [2, 0, 1], [2, 0, 1]]),
     np.array([1, 1, 1, 2, 2, 2, 3, 3, 3]))
])
def test_find_consensus(assignments, clusters):
    """Test finding consensus clustering."""
    assert np.all(modularity.find_consensus(assignments) == clusters)


def test_dummyvar():
    """Test generation of dummy variables."""
    # generate small example dummy variable code
    out = modularity._dummyvar(np.array([1, 1, 2, 3, 3]))
    assert np.all(out == np.array([[1, 0, 0],
                                   [1, 0, 0],
                                   [0, 1, 0],
                                   [0, 0, 1],
                                   [0, 0, 1]]))

    allones = np.array([1, 1, 1, 1, 1, 1, 1, 1])
    assert np.all(modularity._dummyvar(allones) == allones)


def test_zrand():
    """Test calculation of zrand."""
    # make the same two-group community assignments (with different labels)
    label = np.ones((100, 1))
    X, Y = np.vstack((label, label * 2)), np.vstack((label * 2, label))
    # compare
    assert modularity.zrand(X, Y) == modularity.zrand(X, Y[::-1])
    random = rs.choice([0, 1], size=X.shape)
    assert modularity.zrand(X, Y) > modularity.zrand(X, random)
    assert modularity.zrand(X, Y) == modularity.zrand(X[:, 0], Y[:, 0])


def test_zrand_partitions():
    """Test calculation of zrand for partitions."""
    # make random communities
    comm = rs.choice(range(6), size=(10, 100))
    all_diff = modularity._zrand_partitions(comm)
    all_same = modularity._zrand_partitions(np.repeat(comm[:, [0]], 10, axis=1))

    # partition of labels that are all the same should have higher average
    # zrand and lower stdev zrand
    assert np.nanmean(all_same) > np.nanmean(all_diff)
    assert np.nanstd(all_same) < np.nanstd(all_diff)



"""Utility functions for working with interfaces."""

PARCIGNORE = [
    "unknown",
    "corpuscallosum",
    "Background+FreeSurfer_Defined_Medial_Wall",
    "???",
    "Unknown",
    "Medial_wall",
    "Medial wall",
    "medial_wall",
]
